# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

from contextlib import contextmanager
import logging
import sys

import click

try:
    import odoo
    from odoo.api import Environment
    Registry = odoo.modules.registry.Registry
    odoo_bin = 'odoo'
except ImportError:
    # Odoo < 10
    try:
        import openerp as odoo
        from openerp.api import Environment
        Registry = odoo.modules.registry.RegistryManager
        odoo_bin = 'openerp-server'
    except ImportError:
        raise ImportError("No module named odoo nor openerp")

_logger = logging.getLogger(__name__)


def _fix_logging(series):
    if series < 9:
        handlers = logging.getLogger().handlers
        if handlers and len(handlers) == 1:
            handler = handlers[0]
            if isinstance(handler, logging.StreamHandler):
                if handler.stream is sys.stdout:
                    handler.stream = sys.stderr


def parse_config(config=None, database=None, log_level=None, logfile=None):
    series = odoo.release.version_info[0]

    odoo_args = []
    if config:
        odoo_args.extend(['-c', config])
    if database:
        odoo_args.extend(['-d', database])
    if log_level:
        odoo_args.extend(['--log-level', log_level])
    if logfile:
        odoo_args.extend(['--logfile', logfile])
    # see https://github.com/odoo/odoo/commit/b122217f74
    odoo.tools.config['load_language'] = None
    odoo.tools.config.parse_config(odoo_args)
    _fix_logging(series)
    odoo.cli.server.report_configuration()


@contextmanager
def OdooEnvironment(config=None, database=None, log_level=None, logfile=None,
                    rollback=False):
    parse_config(config, database, log_level, logfile)

    db_name = odoo.tools.config['db_name']
    if not db_name:
        raise click.ClickException("No database name found. Please provide "
                                   "one with the -d option or the odoo "
                                   "configuration file.")

    with Environment.manage():
        registry = odoo.registry(db_name)
        try:
            with registry.cursor() as cr:
                uid = odoo.SUPERUSER_ID
                try:
                    ctx = Environment(cr, uid, {})['res.users'].context_get()
                except Exception as e:
                    ctx = {'lang': 'en_US'}
                    # this happens, for instance, when there are new
                    # fields declared on res_partner which are not yet
                    # in the database (before -u)
                    _logger.warn(
                        "Could not obtain a user context, continuing "
                        "anyway with a default context. Error was: %s",
                        e
                    )
                env = Environment(cr, uid, ctx)
                cr.rollback()
                try:
                    yield env
                    if rollback:
                        cr.rollback()
                    else:
                        cr.commit()
                except:  # noqa
                    _logger.exception("Exception in OdooEnvironment")
                    raise
        finally:
            Registry.delete(db_name)
            odoo.sql_db.close_db(db_name)



import sys

env = env  # noqa


op = ''
if len(sys.argv) > 1:
    op = sys.argv[1]

env['ir.config_parameter'].set_param('testparam', 'testvalue')
if op == 'commit':
    env.cr.commit()
elif op == 'rollback':
    env.cr.rollback()
elif op == 'raise':
    raise RuntimeError("testerror")
else:
    pass



# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

import functools
import logging
import runpy
import sys

import click

from . import console
from .env import OdooEnvironment, parse_config, odoo

_logger = logging.getLogger(__name__)


def _remove_click_option(func, name):
    for o in func.__click_params__:
        if o.name == name:
            func.__click_params__.remove(o)
            return


def env_options(default_log_level='info', with_rollback=True,
                with_database=True, database_required=True):
    def inner(func):
        @click.option('--config', '-c', envvar=['ODOO_RC', 'OPENERP_SERVER'],
                      type=click.Path(exists=True, dir_okay=False),
                      help="Specify the Odoo configuration file. Other "
                           "ways to provide it are with the ODOO_RC or "
                           "OPENERP_SERVER environment variables, "
                           "or ~/.odoorc (Odoo >= 10) "
                           "or ~/.openerp_serverrc.")
        @click.option('--database', '-d', envvar=['PGDATABASE'],
                      required=database_required,
                      help="Specify the database name.")
        @click.option('--log-level',
                      default=default_log_level,
                      show_default=True,
                      help="Specify the logging level. Accepted values depend "
                           "on the Odoo version, and include debug, info, "
                           "warn, error.")
        @click.option('--logfile',
                      type=click.Path(dir_okay=False),
                      help="Specify the log file.")
        @click.option('--rollback', is_flag=True,
                      help="Rollback the transaction even if the script "
                           "does not raise an exception. Note that if the "
                           "script itself commits, this option has no effect. "
                           "This is why it is not named dry run. This option "
                           "is implied when an interactive console is "
                           "started.")
        @functools.wraps(func)
        def wrapped(config, log_level, logfile, database=None, rollback=False,
                    *args, **kwargs):
            if database:
                with OdooEnvironment(
                    config=config,
                    database=database,
                    log_level=log_level,
                    logfile=logfile,
                    rollback=rollback,
                ) as env:
                    return func(env, *args, **kwargs)
            else:
                parse_config(config, None, log_level, logfile)
                with odoo.api.Environment.manage():
                    return func(None, *args, **kwargs)
        if not with_database:
            _remove_click_option(wrapped, 'database')
        if not with_rollback or not with_database:
            _remove_click_option(wrapped, 'rollback')
        return wrapped
    return inner


@click.command(help="Execute a python script in an initialized Odoo "
                    "environment. The script has access to a 'env' global "
                    "variable which is an odoo.api.Environment "
                    "initialized for the given database. If no script is "
                    "provided, the script is read from stdin or an "
                    "interactive console is started if stdin appears "
                    "to be a terminal.")
@env_options(database_required=False)
@click.option('--interactive/--no-interactive', '-i',
              help="Inspect interactively after running the script.")
@click.option('--shell-interface',
              help="Preferred shell interface for interactive mode. Accepted "
                   "values are ipython, ptpython, bpython, python. If not "
                   "provided they are tried in this order.")
@click.argument('script', required=False,
                type=click.Path(exists=True, dir_okay=False))
@click.argument('script-args', nargs=-1)
def main(env, interactive, shell_interface, script, script_args):
    global_vars = {'env': env}
    if script:
        sys.argv[1:] = script_args
        global_vars = runpy.run_path(
            script, init_globals=global_vars, run_name='__main__')
    if not script or interactive:
        if console._isatty(sys.stdin):
            if not env:
                _logger.info("No environment set, use `-d dbname` to get one.")
            console.Shell.interact(global_vars, shell_interface)
            if env:
                env.cr.rollback()
        else:
            sys.argv[:] = ['']
            global_vars['__name__'] = '__main__'
            exec(sys.stdin.read(), global_vars)



# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

from __future__ import print_function
import os
import subprocess
import textwrap

import click
from click.testing import CliRunner
import psycopg2
import pytest

import click_odoo
from click_odoo import OdooEnvironment
from click_odoo.cli import main
from click_odoo import console
from click_odoo import odoo, odoo_bin

here = os.path.abspath(os.path.dirname(__file__))

# This hack is necessary because the way CliRunner patches
# stdout is not compatible with the Odoo logging initialization
# mechanism. Logging is therefore tested with subprocesses.
odoo.netsvc._logger_init = True


def _init_odoo_db(dbname):
    subprocess.check_call([
        'createdb', dbname,
    ])
    subprocess.check_call([
        odoo_bin,
        '-d', dbname,
        '--stop-after-init',
    ])


def _drop_db(dbname):
    subprocess.check_call([
        'dropdb', dbname,
    ])


@pytest.fixture(scope='session')
def odoodb():
    dbname = 'click-odoo-test-' + odoo.release.version.replace('.', '-')
    _init_odoo_db(dbname)
    try:
        yield dbname
    finally:
        _drop_db(dbname)


def test_odoo_env(odoodb):
    with OdooEnvironment(database=odoodb) as env:
        admin = env['res.users'].search([('id', '=', 1)])
        assert len(admin) == 1


def test_click_odoo(odoodb):
    """ Test simple access to env in script """
    script = os.path.join(here, 'scripts', 'script1.py')
    cmd = [
        'click-odoo',
        '-d', odoodb,
        script
    ]
    result = subprocess.check_output(cmd, universal_newlines=True)
    assert result == 'admin\n'


def test_cli_runner(odoodb):
    """ Test simple access to env in script (through click CliRunner) """
    script = os.path.join(here, 'scripts', 'script1.py')
    runner = CliRunner()
    result = runner.invoke(main, [
        '-d', odoodb,
        script
    ])
    assert result.exit_code == 0
    assert result.output == 'admin\n'


def test_click_odoo_args(odoodb):
    """ Test sys.argv in script """
    script = os.path.join(here, 'scripts', 'script2.py')
    cmd = [
        'click-odoo',
        '-d', odoodb,
        '--',
        script,
        'a', '-b', '-d',
    ]
    result = subprocess.check_output(cmd, universal_newlines=True)
    assert result == textwrap.dedent("""\
        sys.argv = {} a -b -d
        __name__ = __main__
    """.format(script))


def test_click_odoo_shebang(odoodb):
    """ Test simple access to env in script with click-odoo shebang """
    script = os.path.join(here, 'scripts', 'script1.py')
    cmd = [
        script,
        '-d', odoodb,
    ]
    result = subprocess.check_output(cmd, universal_newlines=True)
    assert result == 'admin\n'


def test_click_odoo_shebang_args(odoodb):
    """ Test script arguments (with click-odoo shebang) """
    script = os.path.join(here, 'scripts', 'script2.py')
    cmd = [
        script,
        '-d', odoodb,
        '--',
        'a', '-b', '-d',
    ]
    result = subprocess.check_output(cmd, universal_newlines=True)
    assert result == textwrap.dedent("""\
        sys.argv = {} a -b -d
        __name__ = __main__
    """.format(script))


def test_interactive_no_script(mocker, odoodb):
    mocker.patch.object(console.Shell, 'ipython')
    mocker.patch.object(console.Shell, 'python')
    mocker.patch.object(console, '_isatty', return_value=True)

    runner = CliRunner()
    result = runner.invoke(main, [
        '-d', odoodb,
    ])
    assert result.exit_code == 0
    assert console.Shell.ipython.call_count == 1
    assert console.Shell.python.call_count == 0


def test_interactive_no_script_preferred_shell(mocker, odoodb):
    mocker.patch.object(console.Shell, 'ipython')
    mocker.patch.object(console.Shell, 'python')
    mocker.patch.object(console, '_isatty', return_value=True)

    runner = CliRunner()
    result = runner.invoke(main, [
        '-d', odoodb,
        '--shell-interface=python',
    ])
    assert result.exit_code == 0
    assert console.Shell.ipython.call_count == 0
    assert console.Shell.python.call_count == 1


def test_logging_stderr(capfd, odoodb):
    script = os.path.join(here, 'scripts', 'script3.py')
    cmd = [
        'click-odoo',
        '-d', odoodb,
        '--',
        script,
    ]
    subprocess.check_call(cmd)
    out, err = capfd.readouterr()
    assert not out
    assert "Modules loaded" in err
    assert "hello from script3" in err


def test_logging_logfile(tmpdir, capfd, odoodb):
    script = os.path.join(here, 'scripts', 'script3.py')
    logfile = tmpdir.join('mylogfile')
    cmd = [
        'click-odoo',
        '-d', odoodb,
        '--logfile', str(logfile),
        '--',
        script,
    ]
    subprocess.check_call(cmd)
    out, err = capfd.readouterr()
    assert not out
    logcontent = logfile.read()
    assert "Modules loaded" in logcontent
    assert "hello from script3" in logcontent


def tests_env_options(odoodb):
    @click.command()
    @click_odoo.env_options()
    def testcmd(env):
        login = env['res.users'].search([('id', '=', 1)]).login
        click.echo("login={}".format(login))

    runner = CliRunner()
    result = runner.invoke(testcmd, [
        '-d', odoodb,
    ])
    assert result.exit_code == 0
    assert 'login=admin\n' in result.output


def _cleanup_testparam(dbname):
    with psycopg2.connect(dbname=dbname) as conn:
        with conn.cursor() as cr:
            cr.execute("DELETE FROM ir_config_parameter "
                       "WHERE key='testparam'")
            conn.commit()
    conn.close()


def _assert_testparam_present(dbname, expected):
    with psycopg2.connect(dbname=dbname) as conn:
        with conn.cursor() as cr:
            cr.execute("SELECT value FROM ir_config_parameter "
                       "WHERE key='testparam'")
            r = cr.fetchall()
            assert len(r) == 1
            assert r[0][0] == expected
    conn.close()


def _assert_testparam_absent(dbname):
    with psycopg2.connect(dbname=dbname) as conn:
        with conn.cursor() as cr:
            cr.execute("SELECT value FROM ir_config_parameter "
                       "WHERE key='testparam'")
            r = cr.fetchall()
            assert len(r) == 0
    conn.close()


def test_write_commit(odoodb):
    """ test commit in script """
    _cleanup_testparam(odoodb)
    script = os.path.join(here, 'scripts', 'script4.py')
    cmd = [
        'click-odoo',
        '-d', odoodb,
        '--',
        script, 'commit'
    ]
    subprocess.check_call(cmd)
    _assert_testparam_present(odoodb, 'testvalue')


def test_write_rollback(odoodb):
    """ test rollback in script """
    _cleanup_testparam(odoodb)
    script = os.path.join(here, 'scripts', 'script4.py')
    cmd = [
        'click-odoo',
        '-d', odoodb,
        '--',
        script, 'rollback'
    ]
    subprocess.check_call(cmd)
    _assert_testparam_absent(odoodb)


def test_write_defaulttx(odoodb):
    """ test click-odoo commits itself """
    _cleanup_testparam(odoodb)
    script = os.path.join(here, 'scripts', 'script4.py')
    cmd = [
        'click-odoo',
        '-d', odoodb,
        '--',
        script
    ]
    subprocess.check_call(cmd)
    _assert_testparam_present(odoodb, 'testvalue')


def test_write_interactive_defaulttx(mocker, odoodb):
    """ test click-odoo rollbacks in interactive mode """
    mocker.patch.object(console.Shell, 'python')
    mocker.patch.object(console, '_isatty', return_value=True)

    _cleanup_testparam(odoodb)
    runner = CliRunner()
    script = os.path.join(here, 'scripts', 'script4.py')
    cmd = [
        '-d', odoodb,
        '--interactive',
        '--',
        script
    ]
    result = runner.invoke(main, cmd)
    assert result.exit_code == 0
    _assert_testparam_absent(odoodb)


def test_write_stdin_defaulttx(odoodb):
    _cleanup_testparam(odoodb)
    script = os.path.join(here, 'scripts', 'script4.py')
    cmd = [
        'click-odoo',
        '-d', odoodb,
        '<', script
    ]
    subprocess.check_call(' '.join(cmd), shell=True)
    _assert_testparam_present(odoodb, 'testvalue')


def test_write_raise(tmpdir, capfd, odoodb):
    """ test nothing is committed if the script raises """
    _cleanup_testparam(odoodb)
    script = os.path.join(here, 'scripts', 'script4.py')
    logfile = tmpdir.join('mylogfile')
    cmd = [
        'click-odoo',
        '-d', odoodb,
        '--logfile', str(logfile),
        '--',
        script, 'raise'
    ]
    r = subprocess.call(cmd)
    assert r != 0
    logcontent = logfile.read()
    assert "testerror" in logcontent
    our, err = capfd.readouterr()
    assert "testerror" in err
    _assert_testparam_absent(odoodb)


def test_env_cache(odoodb):
    """ test a new environment does not reuse cache """
    _cleanup_testparam(odoodb)
    with OdooEnvironment(database=odoodb) as env:
        env['ir.config_parameter'].set_param('testparam', 'testvalue')
        value = env['ir.config_parameter'].get_param('testparam')
        assert value == 'testvalue'
        env.cr.commit()
    _assert_testparam_present(odoodb, 'testvalue')
    _cleanup_testparam(odoodb)
    _assert_testparam_absent(odoodb)
    with OdooEnvironment(database=odoodb) as env:
        value = env['ir.config_parameter'].get_param('testparam')
        assert not value



#!/usr/bin/env python
import os
import subprocess
import sys


odoo_series = sys.argv[1]
odoo_dir = sys.argv[2]


def odoo_installed():
    try:
        import odoo  # noqa
        return True
    except ImportError:
        # odoo < 10
        try:
            import openerp  # noqa
            return True
        except ImportError:
            # odoo not installed
            return False


def odoo_cloned():
    return os.path.isdir(os.path.join(odoo_dir, '.git'))


def clone_odoo():
    subprocess.check_call([
        'git', 'clone',
        '--depth=1',
        '-b', odoo_series,
        'https://github.com/odoo/odoo',
        odoo_dir,
    ])


def install_odoo():
    subprocess.check_call([
        'pip', 'install',
        '-e', odoo_dir,
    ])


def main():
    if not odoo_installed():
        if not odoo_cloned():
            clone_odoo()
        install_odoo()


if __name__ == '__main__':
    main()



# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

import os
from setuptools import setup


here = os.path.abspath(os.path.dirname(__file__))

long_description = []
with open(os.path.join('README.rst')) as f:
    long_description.append(f.read())
with open(os.path.join('CHANGES.rst')) as f:
    long_description.append(f.read())


setup(
    name='click-odoo',
    description='Beautiful, robust CLI for Odoo',
    long_description='\n'.join(long_description),
    use_scm_version=True,
    packages=[
        'click_odoo',
    ],
    include_package_data=True,
    setup_requires=[
        'setuptools_scm',
    ],
    install_requires=[
        'click',
    ],
    license='LGPLv3+',
    author='ACSONE SA/NV',
    author_email='info@acsone.eu',
    url='http://github.com/acsone/click-odoo',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: '
        'GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Framework :: Odoo',
    ],
    entry_points='''
        [console_scripts]
        click-odoo=click_odoo.cli:main
    ''',
)



# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

from .cli import env_options  # noqa
from .env import OdooEnvironment  # noqa
from .env import odoo  # noqa
from .env import odoo_bin  # noqa
